/*
 * file.cc: Disk file class for indigo,
 * a program to insert index entries into LaTeX documents
 *
 * Version 1.1d
 * Last Change: 18 May, 2013
 *
 *
 * Copyright (C) 2009, 2010, 2013
 *
 * Andrew D. Hwang
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * indigo is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * indigo is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pglatex; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <iostream>
#include <vector>
#include <string>
#include <fstream>

#include "utils.h"
#include "line.h"
#include "folio.h"
#include "index_entry.h"
#include "logger.h"

#include "file.h"

/*
 * Input file must contain exactly one "\begin{document}" line.
 *
 * At end of constructor:
 * m_lines contains the input file,
 * m_bookmarks contains line numbers of the page separators,
 * m_pageids contains string representations of folio numbers,
 * m_parts contains bookmark indices of first Roman- and Arabic-numbered pages,
 * m_roman_lo <= n <= m_roman_hi is a valid page,
 * m_arabic_lo <= n <= m_arabic_hi is a valid page.
 */
file::file(FILENAME filename, logger& error_mgr)
  : m_in_body(false)
{
  std::ifstream input(filename.c_str());
  if (!input)
    error("Can't open file \"" + filename + "\"");

  // else read in lines, get separators
  std::string linebuf;
  unsigned int lineno(0); // current line number

  while (getline(input, linebuf))
    {
      // update "in_body" status
      if (!m_in_body)
	m_in_body = begin_document(linebuf);

      else
	m_in_body = !end_document(linebuf);

      // process line
      if(m_in_body)
	{
	  std::string pgnum(is_page_separator(linebuf));

	  if (pgnum != "") // got a separator
	    {
	      m_bookmarks.push_back(lineno);
	      m_pageids.push_back(pgnum);
	    }
	}

      ++lineno;
      m_lines.push_back(linebuf);
    }

  input.close();

  if (m_bookmarks.size() == 0) // No separators, fatal error
    error("Input file " + filename + " contains no page separators.");

  // else
  // add final "fake" separator
  m_bookmarks.push_back(m_lines.size());
  // add 1 to last pageid, convert back to a string
  m_pageids.push_back(rep(1 + to_pgnum(m_pageids.at(m_pageids.size() - 1))));

  // Find start and end of unnumbered, Roman-, and Arabic-numbered pages:
  // initialize m_parts = {n0, n1, n2, n3} so that
  // bookmarks(n0) = line number of first separator
  // bookmarks(n1) = line number of first Roman separator
  // bookmarks(n2) = line number of first Arabic separator
  // bookmarks(n3) = line number one past last line of disk file
  unsigned int ct(0);
  m_parts.push_back(ct);
  while (ct < m_bookmarks.size() && m_pageids.at(ct)=="0") // unnumbered matter
    ++ct;
  m_parts.push_back(ct); // bookmark to first (Roman-)numbered page


  while (ct < m_bookmarks.size() && is_roman(m_pageids.at(ct)))
    ++ct;
  m_parts.push_back(ct); // bookmark first Arabic-numbered page


  while (ct < m_bookmarks.size() && is_arabic(m_pageids.at(ct)))
    ++ct;

  if (ct != m_bookmarks.size())
    error_mgr
      .err(0, "Invalid numbering structure in LaTeX source:")
      .err(0, "Expected unnumbered, Roman-, and Arabic-numbered pages.");

  else
    {
      m_parts.push_back(ct-1); // m_bookmarks.at(m_parts.at(3)) = last page + 1

      // check for consecutive page numbers
      if (m_parts.at(1) < m_parts.at(2)) // found some roman-numbered pages
	{
	  ct = m_parts.at(1); // re-use index
	  m_roman_lo = to_pgnum(m_pageids.at(ct));

	  while (ct + 1 < m_parts.at(2))
	    {
	      if (to_pgnum(m_pageids.at(ct + 1)) !=
		  to_pgnum(m_pageids.at(ct)) + 1)
		error_mgr.err(0, "Page " + m_pageids.at(ct + 1)
			      + " follows page " + m_pageids.at(ct));

	      ++ct;
	    }

	  m_roman_hi = to_pgnum(m_pageids.at(ct));
	}

      else
	{
	  m_roman_lo = m_roman_hi = 0;
	}

      ct = m_parts.at(2);
      if (m_parts.at(2) < m_parts.at(3)) // found some arabic-numbered pages
	{
	  m_arabic_lo = to_pgnum(m_pageids.at(ct));

	  while (ct + 1 < m_parts.at(3))
	    {
	      if (to_pgnum(m_pageids.at(ct + 1)) !=
		  to_pgnum(m_pageids.at(ct)) + 1)
		error_mgr.err(0, "Page " + m_pageids.at(ct + 1)
			      + " follows page " + m_pageids.at(ct));

	      ++ct;
	    }

	  m_arabic_hi = to_pgnum(m_pageids.at(ct));
	}

      else
	{
	  m_arabic_lo = m_arabic_hi = 0;
	}
    }

  if (m_roman_lo == 0)
    {
      if (m_arabic_lo == 0)
	error_mgr.err(0, "No numbered material found.");

      else
	error_mgr.warn("No Roman-numbered material found.");
    }

  else if (m_arabic_lo == 0)
    error_mgr.warn("No Arabic-numbered material found.");
}


void file::mark(const folio& pg, const index_entry& entry)
{
  if (pg.pgnum() == 0) // cross-reference? -> start of first Arabic page
    {
      m_lines.at(m_bookmarks.at(m_parts.at(2)))
	.mark(entry.replace() + pg.suffix()); // force
    }

  else if (is_arabic(pg.pgtag()))
    {
      unsigned int lo(m_bookmarks.at(m_parts.at(2)+pg.pgnum() - m_arabic_lo));
      unsigned int hi(m_bookmarks.at(m_parts.at(2)+pg.pgnum() - m_arabic_lo+1));
      mark(lo, hi, pg, entry);
    }

  else if (is_roman(pg.pgtag()))
    {
      unsigned int lo(m_bookmarks.at(m_parts.at(1)+pg.pgnum() - m_roman_lo));
      unsigned int hi(m_bookmarks.at(m_parts.at(1)+pg.pgnum() - m_roman_lo+1));
      mark(lo, hi, pg, entry);
    }
  // else fail silently; check_ranges should have caught any problems
}


// flag an error unless every page is pglist is in range
void file::check_ranges(const std::list<folio>& pglist, logger& error_mgr) const
{
  for (std::list<folio>::const_iterator p = pglist.begin();
       p != pglist.end(); ++p)
    if ((*p).pgnum() == 0) // cross-reference
      {
	if (m_parts.at(2) == m_parts.at(3)) // nowhere to place it?
	  error_mgr.err(0, "Cross-reference, but no Arabic-numbered pages.");
      }

    else if (is_arabic((*p).pgtag()))
      {
	if (m_arabic_hi < (*p).pgnum() || (*p).pgnum() < m_arabic_lo)
	  error_mgr.err(0, "Index entry requested on non-existent page "
			+ (*p).pgtag() + ".");
      }

    else if (is_roman((*p).pgtag()))
      {
	if (m_roman_hi < (*p).pgnum() || (*p).pgnum() < m_roman_lo)
	  error_mgr.err(0, "Index entry requested on non-existent page "
			+ (*p).pgtag() + ".");
      }

    else
      error_mgr.err(0, "Index entry requested at invalid page " + (*p).pgtag());
}


void file::write(FILENAME filename) const
{
  std::ofstream output(filename.c_str());
  if (!output)
    error("Can't open file \"" + filename + "\"");

  for (std::vector<line>::const_iterator lp = m_lines.begin();
       lp != m_lines.end(); ++lp)
    (*lp).write(output);

  output.close();
}


// private function; pass starting and ending line numbers to check
void file::mark(unsigned int lo, unsigned int hi,
		const folio& pg, const index_entry& entry)
{
  std::string tag(entry.replace() + pg.suffix());

  unsigned int matches(0);
  if (!pg.reverse())
    {
      unsigned int ct(lo);
      while (matches == 0 && ct != hi)
	{
	  matches += m_lines.at(ct).mark(entry.search(), tag);
	  ++ct;
	}

      // If no match, place fallback entry just after page separator
      if (matches == 0)
	m_lines.at(lo).mark(tag); // force
    }

  else
    {
      unsigned int ct(hi);
      while (matches == 0 && ct != lo)
	{
	  --ct;
	  matches += m_lines.at(ct).mark(entry.search(), tag);
	}

      // If no match, place fallback entry just before page separator
      if (matches == 0)
	m_lines.at(hi - 1).mark(tag); // force
    }
}
