/*
 * index_line.cc: Class representing an index line for indigo,
 * a program to insert index entries into LaTeX documents
 *
 * Version 1.1c
 * Last Change: 09 March, 2011
 *
 *
 * Copyright (C) 2009, 2010, 2011
 *
 * Andrew D. Hwang (DP: adhere)
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * indigo is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * indigo is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pglatex; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <list>
#include <string>

#include "folio.h"
#include "index_entry.h"

#include "utils.h"
#include "logger.h"

#include "index_line.h"

const std::string nsep(" ,;."); // number-list token separators

const std::string numeral_characters("1234567890-cilvxCILVX");

// constructor classifies lines and parses page numbers/ranges
index_line::index_line(unsigned int lineno,
		       const std::string& linedata,
		       logger& error_mgr)
  : m_number(lineno), m_type(BLANK)
{
  // remove and store LaTeX comment
  std::string data(strip_comment(linedata, m_comment));

  // count leading spaces
  int spaces(0);
  while (data.substr(0,1) == " ")
    {
      data.erase(0,1);
      ++spaces;
    }

  // \t, multiple spaces -> single space (applies to user's replacement text)
  data = normalize_space(data);

  // coarsely classify input
  if (0 < data.size()) // not all white space
    {
      if(is_number(read_next_token(data, nsep+"-")))
	m_type = FOLIOS; // parse line numbers later


      // remove LaTeX macros where needed; **** LaTeX-specific code
      else if(read_next_token(data, " [") == "\\item")
	{
	  get_next_token(data, " ["); // strip off leading \item, etc.
	  m_type = ITEM;
	}

      else if(spaces == 2)
	m_type = SUBITEM;

      else if(read_next_token(data, " [") == "\\subitem")
	{
	  get_next_token(data, " [");
	  m_type = SUBITEM;
	}

      else if(spaces == 4)
	m_type = SUBSUBITEM;

      else if(read_next_token(data, " [") == "\\subsubitem")
	{
	  get_next_token(data, " [");
	  m_type = SUBSUBITEM;
	}

      else if (spaces == 0)
	m_type = ITEM;

      else
	{
	  m_type = ERROR;
	  m_comment = linedata; // store for posterity
	  error_mgr.err(lineno,
			"Doesn't begin with 0, 2, or 4 spaces, or a number.");
	}
    }
  else // BLANK (includes comment lines)
    m_comment = linedata; // store for posterity


  // Get search and replacement text
  if (m_type == ITEM || m_type == SUBITEM || m_type == SUBSUBITEM)
    {
      if (read_next_token(data, " ") == "[]")
	get_next_token(data, " ");

      // Explicit replacement string delimiter?
      if (data.find_first_of("#") != std::string::npos)
	{
	  m_replace = get_next_token(data, "#"); // store replacement text
	  data.erase(0, 1); // discard leading "#"
	}

      // replacement text removed; replace "--" with clothed -
      data = normalize_ranges(data);

      // Get text up to next arabic page; include 0, which signifies a xref
      m_search = get_next_token(data, "1234567890");

      // We may have grabbed some roman numerals; restore to folio list
      while (is_roman(read_last_token(m_search, nsep)))
	data = get_last_token(m_search, nsep) + " " + data;

      if (m_replace == "") // no explicit replacement found earlier
	m_replace = m_search;
    }

  else if (m_type == FOLIOS) // replace "--" with clothed -
    data = normalize_ranges(data);


  if (m_type != ERROR) // need to preserve ERRORs verbatim
    {
      // clean up text strings; remove specified leaders and trailers
      m_search  = trim(m_search,  " ,;");
      m_replace = trim(m_replace, " ,;");
      m_comment = trim(m_comment);
    }

  // data now contains a list of pages and ranges, separated by characters
  // in nsep (see top of this file). Ranges are denoted by clothed hyphens.
  if (m_type != BLANK && m_type != ERROR)
    while (data.size() != 0)
      {
	std::string tmp(get_next_token(data, nsep));
	if (tmp.size() == 0)
	  break;

	// Recognize Arabic numbers, Roman numerals
	if (tmp.find_first_not_of(numeral_characters) != std::string::npos)
	  {
	    error_mgr.err(lineno, "Folio token \""+tmp+"\" isn't a number.");
	    continue;
	  }

	else if (tmp.find_first_of("-") != std::string::npos) // range
	  {
	    // we've squeezed out "--"s
	    if (tmp.find_first_of("-") != tmp.find_last_of("-"))
	      {
		error_mgr.err(lineno, "Token \""+tmp+"\" isn't a page range.");
		continue;
	      }

	    else // have a range, add two folios
	      {
		// **** LaTeX-specific code
		m_pages.push_back(folio(get_next_token(tmp, "-"), "|("));
		m_pages.push_back(folio(get_next_token(tmp, "-"), "|)", true));
	      }
	  }

	else
	  m_pages.push_back(folio(tmp));
      }
} // end of index_line constructor

void index_line::merge(index_line arg)
{
  m_pages.splice(m_pages.end(), arg.m_pages);
}

void index_line::make_subitem_of(const index_line& item)
{
  m_search = item.m_search;
  m_replace = item.m_replace + "!" + m_replace; //**** LaTeX-specific code
}

void index_line::add_to_list(std::list<index_entry>& entries) const
{
  // ignore BLANK, ERROR, FOLIOS
  // index_data logic ensures we're only called on [SUB]ITEMs, but play it safe
  if (m_type == ITEM || m_type == SUBITEM || m_type == SUBSUBITEM)
    {
      index_entry val(m_search, m_replace);

      for (std::list<folio>::const_iterator p = m_pages.begin();
	   p != m_pages.end(); ++p)
	val.add_page(*p);

      entries.push_back(val);
    }
}

void index_line::write(std::ofstream& outfile) const
{
  using std::endl;
  // special output for each type
  if (m_type == BLANK)
    {
      if (m_comment.size() != 0)
	outfile << trim(m_comment) << endl;
      return;
    }

  else if (m_type == ITEM)
    outfile << endl;

  else if (m_type == SUBITEM || m_type == FOLIOS)
    outfile << "  ";

  else if (m_type == SUBSUBITEM)
    outfile << "    ";

  else if (m_type == ERROR)
    {
      outfile << "%% **** Line " << m_number << " of original:" << endl
	      << m_comment << endl; // this is posterity
      return;
    }

  if (m_type != FOLIOS)
    {
      if (m_search != m_replace)
	outfile << m_replace << "#";

      outfile << m_search << " ";
    }

  // collect page numbers, trim surrounding spaces, commas
  std::stringstream buf;
  for (std::list<folio>::const_iterator f = m_pages.begin();
       f != m_pages.end(); ++f)
    buf << (*f).log_write();

  outfile << trim(buf.str(), " ,");

  if (m_comment.size() != 0)
    outfile << " " << m_comment;

  outfile << endl;
}

// Make four passes through lines:
// 1. Remove BLANKs and ERRORs
// 2. Join FOLIOs to preceding entry
// 3. Subsume SUBITEMs in preceding ITEM
// 4. Subsume SUBSUBITEMs in preceding SUBITEM
std::list<index_line> rejoin(std::list<index_line> lines, logger& error_mgr)
{
  std::list<index_line>::iterator curr(lines.begin()), prev;

  // First pass: remove BLANK and ERROR lines
  while (curr != lines.end())
    {
      if ((*curr).m_type == BLANK || (*curr).m_type == ERROR)
	curr = lines.erase(curr);

      else
	++curr;
    }

  curr = lines.begin();
  if ((*curr).m_type != ITEM) // mark everything up to first ITEM as ERROR
    {
      while (curr != lines.end() && (*curr).m_type != ITEM)
	{
	  std::stringstream buf;
	  if ((*curr).m_type == FOLIOS)
	    buf << "Page numbers not attached to an index item.";

	  else if ((*curr).m_type == SUBITEM)
	    buf << "Subitem not attached to an index item.";

	  else if ((*curr).m_type == SUBSUBITEM)
	    buf << "Subsubitem not attached to a subitem.";

	  else // Better not see this except after botched recoding
	    error("Internal error: Line types have changed?");

	  error_mgr.err((*curr).m_number, buf.str());
	  (*curr).m_type = ERROR;
	  curr = lines.erase(curr);
	}
    }

  // Either "lines" starts with an ITEM, or is empty.

  // Second pass: group FOLIOS with preceding (SUB)ITEMs
  curr = lines.begin();
  while(curr != lines.end())
    {
      if ((*curr).m_type == FOLIOS)
	{
	  prev = curr;
	  --prev; // inspecting serially, no need to backtrack more than once
	  (*prev).merge(*curr);
	  curr = lines.erase(curr);
	}

      else
	++curr;
    }

  // Third pass: subsume SUBITEMs into preceding ITEM
  curr = lines.begin();
  while(curr != lines.end())
    { 
      if ((*curr).m_type == SUBITEM)
	{
	  prev = curr; // search for previous ITEM
	  do
	    {
	      --prev;
	    } while ((*prev).m_type != ITEM); // skip intervening SUBSUBITEMS

	  // lines.begin() is an ITEM, so success is guaranteed
	  (*curr).make_subitem_of(*prev); // subsume
	}

      ++curr;
    }

  // Fourth pass: subsume SUBSUBITEMs into preceding SUBITEM
  curr = lines.begin();
  while(curr != lines.end())
    { 
      if ((*curr).m_type == SUBSUBITEM)
	{
	  prev = curr; // search for previous SUBITEM
	  do
	    {
	      --prev;
	    } while ((*prev).m_type == SUBSUBITEM);

	  if ((*prev).m_type == SUBITEM)
	    {
	      (*curr).make_subitem_of(*prev);
	      ++curr;
	    }

	  else // presumably we found an ITEM
	    {
	      (*curr).m_type = ERROR;
	      error_mgr.err((*curr).m_number,
			    "Subsubitem not attached to a subitem.");
	      curr = lines.erase(curr);
	    }
	}

      else
	++ curr;
    }

  return lines;
}
