/*
 * utils.h: Utility functions for indigo,
 * a program to insert index entries into LaTeX documents
 *
 * Version 1.1d
 * Last Change: 18 May, 2013
 *
 *
 * Copyright (C) 2009, 2010, 2011, 2013
 *
 * Andrew D. Hwang
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * indigo is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * indigo is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pglatex; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <iostream>
#include <sstream>
#include <algorithm>

#include "utils.h"

using std::string;

// functor for converting strings to lowercase
namespace {
  class ToLower
  {
  public:
    ToLower(std::locale const& loc)
      : m_loc(loc) { }

    char operator() (char c) const
    {
      return std::tolower(c, m_loc);
    }

  private:
    std::locale const& m_loc;
  };

  static const ToLower down(std::locale::classic());
} // end of namespace


//%%%%%%%%%%%%%%%%%%% Global Function Declarations %%%%%%%%%%%%%%%%%%%

//%%%%%%%% Miscellaneous utilities %%%%%%%%
void debug_msg(const string& msg)
{
  std::cerr << msg << std::endl;
}

// Action: Print warning string
void warning(const string& msg)
{
  std::cerr << "Warning: " << msg << std::endl;
}

// Action: Print error string and exit with error status
void error(const string& msg)
{
  std::cerr << "Error: " << msg << std::endl;
  exit(1);
}

// Action: execute system command; arg passed as string, not const char*
int run_cmd(const string& arg)
{
  return std::system(arg.c_str());
}

// remove file, return status
int rm_file(const string& filename)
{
  return run_cmd("rm -f " + filename + " 2> /dev/null");
}

// move file to file-1.bak (etc.), keeping existing numbered backups
int mv_safe(const string& filename, unsigned int version)
{
  std::string curr(version == 0 ? filename : ""); // caller-specified name

  if (version != 0) // we're in a recursive call; adjust current filename
    {
      std::stringstream buf;
      buf << filename << "-" << version << ".bak";
      curr = buf.str();
    }

  std::stringstream obuf;
  obuf << filename << "-" << version+1 << ".bak"; // new filename
  std::string next(obuf.str());

  if (run_cmd("test -f " + next) == 0) // file exists, recurse
    mv_safe(filename, version+1);

  return run_cmd("mv " + curr + " " + next + " 2> /dev/null");
}

// Action: verify filename has specified extension, exit with error if not
std::string check_extension(const string& filename, const string& extn)
{
  unsigned int sz(extn.size());
  unsigned int fsz(filename.size());

  if (fsz <= sz || filename.substr(fsz - sz, sz) != extn)
    {
      error("Input filename must have extension \"" + extn + "\"");
      return ""; // never get here, but avoids compiler warning
    }

  else
    return filename.substr(0, fsz - sz);
}

// Action: Convert int to string
string rep(int n)
{
  std::stringstream buf;
  buf << n;
  return buf.str();
}

// Action: Convert string (including roman numerals containing "ilvx")
unsigned int to_pgnum(std::string arg)
{
  // all digits?
  if (is_arabic(arg))
    return atoi(arg.c_str());

  else
    {
      // check arg is syntactically legitimate and convert to an int
      if (is_roman(arg))
	{
	  int val(0);
	  arg = lowercase(arg);

	  while (arg.substr(0,1) == "c")
	    {
	      val += 100;
	      arg.erase(0,1);
	    }

	  // check for mutually exclusive "one-time" strings
	  if (arg.substr(0, 2) == "xc")
	    {
	      val += 90;
	      arg.erase(0,2);
	    }

	  else if (arg.substr(0,1) == "l")
	    {
	      val += 50;
	      arg.erase(0,1);
	    }

	  else if (arg.substr(0, 2) == "xl")
	    {
	      val += 40;
	      arg.erase(0,2);
	    }

	  // rest of string consists of "ivx"

	  while (arg.substr(0,1) == "x")
	    {
	      val += 10;
	      arg.erase(0,1);
	    }

	  if (arg.substr(0,1) == "v")
	    {
	      val += 5;
	      arg.erase(0,1);
	    }

	  if (arg == "ix")
	    return val + 9;

	  else if (arg == "iv")
	    return val + 4;

	  else 
	    while (arg.substr(0,1) == "i")
	      {
		val += 1;
		arg.erase(0,1);
	      }
	  return val;
	}

      else
	return 0;
    }
}


//%%%%%%%% String manipulation %%%%%%%%

// Action: Convert input to lowercase in std::locale::classic
string lowercase(string s)
{
  std::transform(s.begin(), s.end(), s.begin(), down);
  return s;
}

// Separate off any LaTeX-commented portion, return remainder
string strip_comment(const string& s, string& comment)
{
  string::size_type end(s.find_first_of("%"));
  if (end == string::npos)
    {
      comment = "";
      return s;
    }

  else
    {
      comment = s.substr(end, s.size());
      return s.substr(0, end);
    }
}

// Remove leading and trailing chars (spaces by default)
string trim(const string& s, const std::string& chars)
{
  string::size_type begin(s.find_first_not_of(chars));

  if (begin == string::npos)
    return "";
  else
    return s.substr(begin, s.find_last_not_of(chars) - begin + 1);
}

// Read and remove next token delimited by specified characters
std::string get_next_token(std::string& arg, const std::string& delims)
{
  std::string::size_type first_char(arg.find_first_not_of(delims));
  if (first_char == std::string::npos)
    {
      arg = "";
      return "";
    }

  else
    {
      arg = arg.substr(first_char, arg.size() - first_char);
      std::string::size_type last_char(arg.find_first_of(delims));
      if (last_char == std::string::npos)
	{
	  std::string val(arg);
	  arg = "";
	  return val;
	}

      else
	{
	  std::string val(arg.substr(0, last_char));
	  arg = arg.substr(last_char, arg.size() - last_char);
	  return val;
	}
    }
}

// Read next token delimited by specified characters
std::string read_next_token(const std::string& arg, const std::string& delims)
{
  std::string::size_type first_char(arg.find_first_not_of(delims));
  if (first_char == std::string::npos)
    return "";

  else
    {
      std::string tmp(arg.substr(first_char, arg.size() - first_char));
      std::string::size_type last_char(tmp.find_first_of(delims));
      if (last_char == std::string::npos)
	return tmp;

      else
	return tmp.substr(0, last_char);
    }
}

// Read and remove last token delimited by specified characters
std::string get_last_token(std::string& arg, const std::string& delims)
{
  std::string::size_type last_char(1+arg.find_last_not_of(delims));
  if (last_char == std::string::npos)
    {
      arg = "";
      return "";
    }

  else
    {
      arg = arg.substr(0, last_char);
      std::string::size_type first_char(1+arg.find_last_of(delims));
      if (first_char == std::string::npos)
	{
	  std::string val(arg);
	  arg = "";
	  return val;
	}

      else
	{
	  std::string val(arg.substr(first_char, arg.size() - first_char));
	  arg = arg.substr(0, first_char);
	  return val;
	}
    }
}


// Read last token delimited by specified characters
std::string read_last_token(const std::string& arg, const std::string& delims)
{
  std::string::size_type last_char(1+arg.find_last_not_of(delims));
  if (last_char == std::string::npos)
    return "";

  else
    {
      std::string tmp(arg.substr(0, last_char));
      std::string::size_type first_char(1+tmp.find_last_of(delims));
      if (first_char == std::string::npos)
	return tmp;

      else
	return tmp.substr(first_char, arg.size() - first_char);
    }
}

std::string read_newline(const std::string& arg)
{
  std::string::size_type eod(arg.find_last_not_of("\r\n")); // end of data
  if (eod == std::string::npos)
    return arg;

  else
    return arg.substr(eod + 1, arg.size() - eod);
}

std::string replace_all(std::string arg,
			const std::string& from, const std::string& to)
{
  std::string::size_type pos(arg.find(from));
  while (pos != std::string::npos)
    {
      arg.replace(pos, from.size(), to);
      pos = arg.find(from);
    }
  return arg;
}


// Replace tabs, newlines, and multiple spaces by a single space
std::string normalize_space(std::string arg)
{
  arg = replace_all(arg, "\r", " ");
  arg = replace_all(arg, "\t", " ");
  arg = replace_all(arg, "\n", " ");
  arg = replace_all(arg, "  ", " "); // compress consecutive spaces

  // Remove leading and trailing space
  if (arg.substr(0, 1) == " ")
    arg.erase(0, 1);

  if (arg.size() != 0 && arg.substr(arg.size() - 1, 1) == " ")
    arg.erase(arg.size() - 1, 1);

  return arg;
}

// Remove spaces around page ranges, convert to hyphens
std::string normalize_ranges(std::string arg)
{
  // clothe en-dashes; assuming arg contains no em-dashes
  arg = replace_all(arg, " -- ", "-");
  arg = replace_all(arg, "-- ", "-");
  arg = replace_all(arg, " --", "-");
  arg = replace_all(arg, "--", "-");

  // clothe hyphens
  arg = replace_all(arg, " - ", "-");
  arg = replace_all(arg, "- ", "-");
  return replace_all(arg, " -", "-");
}


//%%%%%%%% Matching criteria %%%%%%%%
// Found \begin{document} on a line by itself?
bool begin_document(const std::string& arg)
{
  return arg.find("begin{document}") != std::string::npos;
}

bool end_document(const std::string& arg)
{
  return arg.find("end{document}") != std::string::npos;
}

// Input: text string
// Output: "": No; otherwise folio number ("0" = unnumbered)
std::string is_page_separator(const std::string& arg)
{
  if (arg.find(".png---") != std::string::npos ||
      arg.find("PageSep") != std::string::npos)
    {
      std::string pg(read_last_token(arg, " -.,;:[](){}%\t\n\r"));

      if (is_number(pg))
	return pg;

      else
	return "0";
    }

  else // Not a separator
    return "";
}

// Found criterion in data (case insensitive)?
bool match(const std::string& data, const std::string& search)
{
  if (data.find(search) == string::npos)
    return false;

  // else check for whole word match
  // pad data front/back with spaces, then
  // look for criterion surrounded by non-alphabetic characters
  string mydata(" " + data + " ");

  string::size_type q(mydata.find(search));

  while(q != string::npos)
    {
      // get characters immediately preceding and following matching substring
      string prev(mydata.substr(q-1,1));
      string next(mydata.substr(q+search.size(),1));

      if (!isalpha(prev.at(0)) && !isalpha(next.at(0)))
	return true;

      // else
      mydata.erase(0, q + search.size() - 1);

      q = mydata.find(search); // next match
    }
  return false;
}

// Roman numeral detection
namespace {
  bool is_roman_i(const std::string& arg)
  {
    if (arg.find_first_not_of("i") == std::string::npos &&
	0 < arg.size() && arg.size() < 4)
      return true;

    return false;
  }

  bool is_roman_iv(const std::string& arg)
  {
    if (arg.find_first_not_of("iv") != std::string::npos ||
	arg.size() == 0 || arg.size() > 4)
      return false;

    if (is_roman_i(arg) || // eliminates empty string
	(arg.size() <= 2 && arg != "vv") ||
	(arg.substr(0, 1) == "v" && is_roman_i(arg.substr(1, arg.size() - 1))))
      return true;

    else
      return false;
  }

  bool is_roman_ivx(const std::string& arg)
  {
    if (arg.find_first_not_of("ivx") != std::string::npos ||
	arg.size() == 0)
      return false;

    if (is_roman_iv(arg) ||
	(arg.find_first_not_of("x") == std::string::npos && arg.size() < 4) ||
	// legitimate "out of order" numerals
	arg == "ix" || arg == "xix" || arg == "xxix" || arg == "xxxix")
      return true;

    std::string::size_type last_x(arg.find_first_not_of("x"));
    if (3 < last_x || !is_roman_iv(arg.substr(last_x, arg.size() - last_x)))
      return false;

    return true;
  }

  bool is_roman_ilvx(const std::string& arg)
  {
    if (arg.find_first_not_of("ilvx") != std::string::npos ||
	arg.size() == 0)
      return false;

    // if we contain "l", we start with "l" or "xl"
    if (arg.substr(0,1) == "l")
      {
	if (arg.size() == 1) // arg == "l"
	  return true;

	else
	  return is_roman_ivx(arg.substr(1,arg.size() - 1));
      }

    else if (arg.substr(0,2) == "xl")
      {
	if (arg.size() == 2 || // arg == "xl"
	    arg == "xlix") // only legitimate "xl..." containing another "x"
	  return true;

	else
	  return is_roman_iv(arg.substr(2,arg.size() - 2));
      }

    else
      return is_roman_ivx(arg);
  }

  bool is_roman_cilvx(const std::string& arg)
  {
    if (arg.find_first_not_of("cilvx") != std::string::npos ||
	arg.size() == 0)
      return false;

    std::string::size_type non_c(arg.find_first_not_of("c"));

    if (non_c == std::string::npos) // all "c"s
      return (arg.size() < 4);

    else if (3 < non_c)
      return false;

    else // arg contains a "c", starts with < 4 "c"s, and is not all "c"s
      {
	if (arg.substr(non_c,2) == "xc") // check for legit "out of order" arg
	  {
	    if (arg.size() == non_c + 2 ||
		(arg.size() < non_c + 5 && arg.substr(non_c,4) == "xcix") ||
		is_roman_iv(arg.substr(non_c + 2,arg.size() - non_c - 2)))
	      return true;

	    else
	      return false;
	  }

	else // no "xc"
	  return is_roman_ilvx(arg.substr(non_c,arg.size() - non_c));
      }
  }
}

bool is_roman(const std::string& arg)
{
  return is_roman_cilvx(lowercase(arg));
}

/*
bool is_Roman(const std::string& arg)
{
  if (arg.find_first_not_of("IVX") != std::string::npos)
    return false;

  return is_roman(lowercase(arg));
}
*/

bool is_arabic(const std::string& arg)
{
  return (arg.find_first_not_of("1234567890") == std::string::npos);
}

bool is_number(const std::string& arg)
{
  return is_arabic(arg) || is_roman(arg);
}
