/*
 * utils.h: Utility functions for indigo,
 * a program to insert index entries into LaTeX documents
 *
 * Version 1.1d
 * Last Change: 18 May, 2013
 *
 *
 * Copyright (C) 2009, 2010, 2011, 2013
 *
 * Andrew D. Hwang
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * indigo is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * indigo is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pglatex; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef DP_INDIGO_UTILS
#define DP_INDIGO_UTILS

#include <string>
#include <sstream>

#include "typedefs.h"
class folio;

//%%%%%%%%%%%%%%%%%%% Global Function Declarations %%%%%%%%%%%%%%%%%%%

//%%%%%%%% Miscellaneous utilities %%%%%%%%

void debug_msg(const std::string& msg);

// Action: Print error string
void warning(const std::string& msg);

// Action: Print error string and exit with error status
void error(const std::string& msg);

// Input: command string
// Return value: exit status of command
// Action: execute system command (N.B. arg passed as string, not const char*)
int run_cmd(const std::string& arg);

// remove file, return status
int rm_file(const std::string&);

// move file to file-1.bak (etc.), keeping existing numbered backups
int mv_safe(const std::string&, unsigned int version=0);

// Action: Verify filename has specified extension;
// Return basename on success, or exit with error
std::string check_extension(const std::string& filename, const std::string& extn);

// Action: Convert int to string
std::string rep(int);

// Convert string representing a positive integer
// (including roman numerals comprising c, i, l, v, x) to int
// Returns 0 if input is neither all digits nor a roman numeral
unsigned int to_pgnum(std::string);

//%%%%%%%% String manipulation %%%%%%%%

// Action: Convert input to lowercase in std::locale::classic
std::string lowercase(std::string);
std::string lowercase(const char*);

// Separate off any LaTeX-commented portion, return remainder
std::string strip_comment(const std::string& s, std::string& comment);

// Remove leading and trailing chars
std::string trim(const std::string& s, const std::string& chars = " ");
// Read and remove the next integer from arg
//folio get_next_number(std::string& arg);

// Read and remove next token delimited by specified characters
std::string get_next_token(std::string& arg, const std::string& delims);
// Read next token delimited by specified characters
std::string read_next_token(const std::string& arg, const std::string& delims);

// Read and remove last token delimited by specified characters
std::string get_last_token(std::string& arg, const std::string& delims);
// Read last token delimited by specified characters
std::string read_last_token(const std::string& arg, const std::string& delims);

// Return trailing newline character(s)
std::string read_newline(const std::string& arg);

// In arg, sequentially replace all instances of from with to.
// Caution: Results may be unexpected if a replacement creates
// a new instance of from.
std::string replace_all(std::string arg,
			const std::string& from, const std::string& to);

// Replace tabs, newlines, and multiple spaces by a single space
std::string normalize_space(std::string);
// Remove spaces around page range hyphens
std::string normalize_ranges(std::string arg);

//////////////////// String classification ////////////////////

//%%%%%%%% Matching criteria %%%%%%%%
// locate start/end of document body
bool begin_document(const std::string& arg);
bool end_document(const std::string& arg);

// Input: text string
// Output: "": No; otherwise folio number ("0" = unnumbered)
std::string is_page_separator(const std::string& arg);

// Found criterion in data (case insensitive)?
bool match(const std::string& data, const std::string& criterion);

bool is_roman(const std::string& arg);
bool is_arabic(const std::string& arg);
bool is_number(const std::string&);
#endif /* DP_INDIGO_UTILS */
